from abc import ABC, abstractmethod
from itertools import groupby

import humanize
from matplotlib.colors import hsv_to_rgb, to_hex
from nicegui import ui

from .kpi_buckets import Month, TimeBucket, Week
from .kpi_logger import KpiLogger, str_to_date


class kpi_page(ABC):

    def __init__(self, kpi_logger: KpiLogger) -> None:
        @ui.page('/kpis')
        def page():
            if self.language == 'de':
                humanize.activate('de')
            else:
                humanize.deactivate()
            with ui.row().style('margin:1em'):
                ui.markdown(f'### {self.title}').style('margin:1.5em;margin-top:-1.2em;')
                toggle = ui.toggle(self.timespans, value=list(self.timespans)[0], on_change=lambda e: show(e.value))
            with ui.row():
                positive_chart = ui.chart({
                    'title': {'text': self.positive_title},
                    'chart': {'type': 'column'},
                    'xAxis': {},
                    'yAxis': {'title': False},
                    'credits': False,
                    'exporting': False,
                })
                negative_chart = ui.chart({
                    'title': {'text': self.negative_title},
                    'chart': {'type': 'column'},
                    'plotOptions': {'series': {'stacking': 'normal'}},
                    'xAxis': {},
                    'yAxis': {'title': False},
                    'colors': [to_hex(hsv_to_rgb((0, 1, i / 100.0))) for i in reversed(range(20, 120, 20))],
                    'credits': False,
                    'exporting': False,
                })
            ui.timer(5, lambda: show(toggle.value))

            def show(num_days: int) -> None:
                if num_days <= 7:
                    time_buckets = kpi_logger.days[-num_days:]
                elif num_days % 7 == 0:
                    weeks = [list(v) for _, v in groupby(kpi_logger.days,
                                                         key=lambda d: str_to_date(d.date).strftime('%y-%W'))][-num_days//7:]
                    time_buckets = [Week.from_buckets(days_in_week) for days_in_week in weeks]
                elif num_days % 30 == 0 and num_days <= 90:
                    months = [list(v) for _, v in groupby(kpi_logger.days,
                                                          key=lambda d: str_to_date(d.date).strftime('%y-%m'))][-num_days//30:]
                    time_buckets = [Month.from_buckets(days_in_month) for days_in_month in months]
                else:
                    raise ValueError(f'Unsupported number of days: {num_days}')

                keys = set(key for day in time_buckets for key in day.incidents if key in self.positives)
                data = {self.positives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
                positive_chart.options['xAxis']['categories'] = [_label(b) for b in time_buckets]
                positive_chart.options['series'] = [{'name': k, 'data': v} for k, v in sorted(data.items())]
                positive_chart.update()

                keys = set(key for day in time_buckets for key in day.incidents if key in self.negatives)
                data = {self.negatives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
                negative_chart.options['xAxis']['categories'] = [_label(b) for b in time_buckets]
                negative_chart.options['series'] = [{'name': k, 'data': v} for k, v in sorted(data.items())]
                negative_chart.update()

    @property
    def language(self) -> str:
        return 'en'

    @property
    @abstractmethod
    def title(self) -> str:
        pass

    @property
    @abstractmethod
    def positive_title(self) -> str:
        pass

    @property
    @abstractmethod
    def negative_title(self) -> str:
        pass

    @property
    @abstractmethod
    def positives(self) -> dict[str, str]:
        pass

    @property
    @abstractmethod
    def negatives(self) -> dict[str, str]:
        pass

    @property
    @abstractmethod
    def timespans(self) -> dict[int, str]:
        pass


def _label(time_bucket: TimeBucket) -> str:
    if isinstance(time_bucket, Month):
        return str_to_date(time_bucket.date).strftime('%b')
    elif isinstance(time_bucket, Week):
        return f'#{str_to_date(time_bucket.date).isocalendar()[1]}'
    else:
        return humanize.naturaldate(str_to_date(time_bucket.date))



from .area import Area
from .area_object_ import AreaObject as area_object
from .obstacle import Obstacle
from .obstacle_object_ import ObstacleObject as obstacle_object
from .path_object_ import PathObject as path_object
from .path_planner import PathPlanner



import abc
import logging
from typing import Optional

from aenum import Enum, auto

from ..event import Event
from .detections import Detections
from .image import Image
from .uploads import Uploads


class Autoupload(Enum, init='value __doc__'):
    """Configures the auto-submitting of images to the Learning Loop"""

    def _generate_next_value_(name: str, start, count, last_values) -> str:
        """uses enum name as value when calling auto()"""
        return name.lower()

    FILTERED = auto(), 'only submit images with novel detections and in an uncertainty range (this is the default)'
    DISABLED = auto(), 'no auto-submitting'
    ALL = auto(), 'submit all images which are run through the detector'


class Detector(abc.ABC):
    """A detector allows detecting objects in images.

    It also holds an upload queue for sending images with uncertain results to an active learning infrastructure like the [Zauberzeug Learning Loop](https://zauberzeug.com/learning-loop.html).
    """

    def __init__(self) -> None:
        self.NEW_DETECTIONS = Event()
        """detection on an image is completed (argument: image)"""
        self.log = logging.getLogger('rosys.detector')

    @abc.abstractmethod
    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED) -> Optional[Detections]:
        return

    @abc.abstractmethod
    async def upload(self, image: Image) -> None:
        pass

    @property
    def uploads(self) -> Uploads:
        return



import os
from typing import Optional

import serial
from nicegui import ui
from nicegui.events import ValueChangeEventArguments

from ... import rosys
from .communication import Communication


class SerialCommunication(Communication):
    """This module implements a communication via a serial device with a given baud rate.

    It contains a list of search paths for finding the serial device.
    """

    search_paths: list[str] = [
        '/dev/tty.SLAB_USBtoUART',
        '/dev/ttyTHS1',
        '/dev/ttyUSB0',
    ]

    def __init__(self, baud_rate: int = 115200) -> None:
        super().__init__()
        self.device_path = self.get_device_path()
        if self.device_path is None:
            raise Exception('No serial port found')
        self.log.debug(f'connecting serial on {self.device_path} with baud rate {baud_rate}')
        self.serial = serial.Serial(self.device_path, baud_rate)
        self.buffer = ''
        self.log_io: bool = False

    @staticmethod
    def is_possible() -> bool:
        return SerialCommunication.get_device_path() is not None

    @staticmethod
    def get_device_path() -> Optional[str]:
        for device_path in SerialCommunication.search_paths:
            if os.path.exists(device_path) and os.stat(device_path).st_gid > 0:
                return device_path

    def connect(self) -> None:
        if not self.serial.isOpen():
            self.serial.open()
            self.log.debug(f'reconnected serial on {self.device_path}')

    def disconnect(self) -> None:
        if self.serial.isOpen():
            self.serial.close()
            self.log.debug(f'disconnected serial on {self.device_path}')

    async def read(self) -> Optional[str]:
        if not self.serial.isOpen():
            return
        s = self.serial.read_all()
        try:
            s = s.decode()
            self.buffer += s
            if '\n' in self.buffer:
                line, self.buffer = self.buffer.split('\r\n', 1)
                if self.log_io:
                    self.log.debug(f'read: {line}')
                return line
        except Exception:
            self.log.exception(f'Could not decode serial data: {s}')

    async def send(self, line: str) -> None:
        if not self.serial.isOpen():
            return
        self.serial.write(f'{line}\n'.encode())
        if self.log_io:
            self.log.debug(f'send: {line}')

    def debug_ui(self) -> None:
        super().debug_ui()

        async def submit_input() -> None:
            await self.send(input.value)
            input.value = ''

        def toggle(e: ValueChangeEventArguments) -> None:
            if e.value:
                self.connect()
                rosys.notify('connected to Lizard')
            else:
                self.disconnect()
                rosys.notify('disconnected from Lizard')

        ui.switch('Serial Communication', value=self.serial.isOpen(), on_change=toggle)
        ui.switch('Serial Logging').bind_value(self, 'log_io')
        input = ui.input('Serial Command', on_change=submit_input)



from functools import wraps

from nicegui import ui


class Track:

    def __init__(self) -> None:
        self.stack: list[str] = []

    def __call__(self, f):
        @wraps(f)
        async def wrap(*args, **kw):
            try:
                self.stack.append(f.__name__)
                return await f(*args, **kw)
            finally:
                if self.stack and self.stack[-1] == f.__name__:
                    self.stack.pop()
        return wrap

    def ui(self) -> ui.label:
        label = ui.label()
        ui.timer(0.5, lambda: label.set_text(' → '.join(self.stack)))
        return label

    def reset(self) -> None:
        self.stack.clear()


track = Track()



import abc
import logging
import uuid
from dataclasses import dataclass, field
from multiprocessing import Process
from multiprocessing.connection import Connection
from typing import Any

from ..geometry import Point, Pose, Spline
from .area import Area
from .delaunay_planner import DelaunayPlanner
from .obstacle_map import Obstacle


@dataclass
class PlannerCommand(abc.ABC):
    id: str = field(init=False, repr=False)
    deadline: float

    def __post_init__(self) -> None:
        self.id = str(uuid.uuid4())


@dataclass
class PlannerSearchCommand(PlannerCommand):
    areas: list[Area]
    obstacles: list[Obstacle]
    start: Pose
    goal: Pose


@dataclass
class PlannerGrowMapCommand(PlannerCommand):
    points: list[Point]


@dataclass
class PlannerTestCommand(PlannerCommand):
    areas: list[Area]
    obstacles: list[Obstacle]
    spline: Spline
    backward: bool = False


@dataclass
class PlannerObstacleDistanceCommand(PlannerCommand):
    areas: list[Area]
    obstacles: list[Obstacle]
    pose: Pose
    backward: bool = False


@dataclass
class PlannerResponse:
    id: str
    deadline: float
    content: Any


class PlannerProcess(Process):

    def __init__(self, connection: Connection, robot_outline: list[tuple[float, float]]) -> None:
        super().__init__()
        self.log = logging.getLogger('rosys.pathplanning.PlannerProcess')
        self.connection = connection
        self.planner = DelaunayPlanner(robot_outline)

    def run(self) -> None:
        while True:
            try:
                cmd = self.connection.recv()
            except EOFError:
                self.log.info('PlannerProcess stopped')
                return
            try:
                if isinstance(cmd, PlannerSearchCommand):
                    self.log.info(cmd)
                    additional_points = [cmd.start.point, cmd.goal.point]
                    self.planner.update_map(cmd.areas, cmd.obstacles, additional_points, cmd.deadline)
                    self.respond(cmd, self.planner.search(cmd.start, cmd.goal))
                if isinstance(cmd, PlannerGrowMapCommand):
                    self.planner.grow_map(cmd.points, cmd.deadline)
                    self.respond(cmd, None)
                if isinstance(cmd, PlannerTestCommand):
                    self.planner.update_map(cmd.areas, cmd.obstacles, [cmd.spline.start, cmd.spline.end], cmd.deadline)
                    self.respond(cmd, bool(self.planner.obstacle_map.test_spline(cmd.spline, cmd.backward)))
                if isinstance(cmd, PlannerObstacleDistanceCommand):
                    self.planner.update_map(cmd.areas, cmd.obstacles, [cmd.pose], cmd.deadline)
                    self.respond(cmd, self.planner.obstacle_map.get_distance(cmd.pose.x, cmd.pose.y, cmd.pose.yaw))
            except Exception as e:
                self.log.exception(f'failed to compute cmd "{cmd}"')
                self.respond(cmd, e)

    def respond(self, cmd: PlannerCommand, content: Any) -> None:
        self.connection.send(PlannerResponse(cmd.id, cmd.deadline, content))



#!/usr/bin/env python
import time

import numpy as np
import pylab as pl
from nicegui import ui
from rosys.pathplanning import plot_tools as pt
from rosys.pathplanning.grid import Grid
from rosys.pathplanning.obstacle_map import ObstacleMap
from rosys.pathplanning.robot_renderer import RobotRenderer
from rosys.pathplanning.steps import Path, Step

grid = Grid((30, 40, 36), (0, 0, 4.0, 3.0))
obstacles = [
    [0.0, 1.0, 2.5, 0.2],
    [1.5, 2.0, 2.5, 0.2],
]
robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)
obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)

path = Path.from_poses([
    [0.7, 0.4, np.deg2rad(5)],
    [1.2, 0.5, 0],
    [1.7, 0.4, np.deg2rad(-15)],
    [2.2, 0.5, 0],
    [2.8, 0.7, np.deg2rad(45)],
    [3.2, 1.0, np.deg2rad(0)],
])

robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)

with ui.plot():
    pt.show_obstacle_map(obstacle_map)
    pl.autoscale(False)
    for step in path:
        pt.plot_robot(robot_renderer, step.target)
    pt.plot_path(path, 'C0', lw=3)

    t = time.time()
    path.smooth(obstacle_map, control_dist=0.2)
    print('%.3f ms' % ((time.time() - t) * 1000))

    pt.plot_path(path, 'C2', lw=3)

    step0 = Step((0.5, 2.9, 0))
    for step in [
        Step((1.0, 2.5, 0), step0),
        Step((1.0, 2.0, 0), step0),
        Step((1.0, 1.5, 1), step0),
        Step((0.4, 1.5, 0), step0),
        Step((0.4, 1.5, 0), step0, backward=True),
        Step((0.1, 2.0, 0), step0, backward=True),
        Step((0.1, 2.5, 0), step0, backward=True),
    ]:
        color = 'C2' if step.is_healthy() else 'C3'
        pt.plot_spline(step.spline, color, backward=step.backward)
        print(step.spline, step.spline.max_curvature(), step.is_healthy())

ui.run()



from .bluetooth import BluetoothHardware
from .bms import Bms, BmsHardware, BmsSimulation
from .bumper import Bumper, BumperHardware, BumperSimulation
from .can import CanHardware
from .communication import Communication, SerialCommunication, WebCommunication
from .estop import EStop, EStopHardware, EStopSimulation
from .expander import ExpanderHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot import Robot, RobotHardware, RobotSimulation
from .robot_brain import RobotBrain
from .serial import SerialHardware
from .wheels import Wheels, WheelsHardware, WheelsSimulation



from __future__ import annotations

import io
from dataclasses import dataclass
from typing import Optional

import PIL.Image
import PIL.ImageDraw

from .detections import Detections


@dataclass(slots=True, kw_only=True)
class ImageSize:
    width: int
    height: int

    @property
    def tuple(self) -> tuple[float, float]:
        return (self.width, self.height)


@dataclass(slots=True, kw_only=True)
class Image:
    camera_id: str
    size: ImageSize
    time: float  # time of recording
    data: Optional[bytes] = None
    detections: Optional[Detections] = None
    is_broken: Optional[bool] = None

    @property
    def id(self) -> str:
        return f'{self.camera_id}/{self.time}'

    @staticmethod
    def create_placeholder(text: str, time: Optional[float] = None, camera_id: Optional[str] = None) -> Image:
        img = PIL.Image.new('RGB', (260, 200), color=(73, 109, 137))
        d = PIL.ImageDraw.Draw(img)
        d.text((img.width/2-len(text)*3, img.height/2-5), text, fill=(255, 255, 255))
        bytesio = io.BytesIO()
        img.save(bytesio, format='PNG')
        image = Image(
            camera_id=camera_id or 'no_cam_id',
            time=time or 0,
            size=ImageSize(width=img.width, height=img.height),
        )
        image.data = bytesio.getvalue()
        return image



from .app_controls_ import AppButton
from .app_controls_ import AppControls as app_controls
from .automation_controls_ import AutomationControls as automation_controls
from .automator import Automator
from .schedule import Schedule



#!/usr/bin/env python3
import numpy as np
from rosys.pathplanning.grid import Grid

grid = Grid((6, 8), (2.0, 1.0, 4.0, 3.0))
x, y = 2.5, 1.0
print(x, y)
row, col = grid.to_grid(x, y)
print(col, row)
x, y = grid.from_grid(row, col)
print(x, y)

grid = Grid((6, 8, 36), (2.0, 1.0, 4.0, 3.0))
x, y, yaw = 2.5, 1.0, np.pi / 2
print(x, y, yaw)
row, col, layer = grid.to_grid(x, y, yaw)
print(col, row, layer)
x, y, yaw = grid.from_grid(row, col, layer)
print(x, y, yaw)



import numpy as np
import pylab as pl


def plot_robot(robot_renderer, pose, color='C0', **kwargs) -> None:
    poly = pl.matplotlib.patches.Polygon(robot_renderer.outline, color='none', ec=color, joinstyle='round', **kwargs)
    triangle = pl.matplotlib.patches.Polygon(
        [[0, 0], [-0.1, -0.05], [-0.1, 0.05]], color='none', ec=color, joinstyle='round', **kwargs)
    pl.gca().add_patch(poly)
    pl.gca().add_patch(triangle)
    move([poly, triangle], pose)


def move(patches, pose) -> None:
    r = pl.matplotlib.transforms.Affine2D().rotate_deg(np.rad2deg(pose[2]))
    t = pl.matplotlib.transforms.Affine2D().translate(pose[0], pose[1])
    for patch in patches:
        patch.set_transform(r + t + pl.gca().transData)


def plot_spline(spline, *args, backward=False, **kwargs) -> None:
    t = np.linspace(0, 1, 20)
    pl.plot(spline.x(t), spline.y(t), *args, alpha=0.5 if backward else 1.0, **kwargs)


def plot_path(path, *args, **kwargs) -> None:
    [plot_spline(step.spline, *args, backward=step.backward, **kwargs) for step in path]


def show_obstacle_map(obstacle_map):
    cmap = pl.matplotlib.colors.ListedColormap([[0, 0, 0, 0.1], 'C3'])
    extent = bbox_to_extent(obstacle_map.grid.bbox)
    pl.imshow(obstacle_map.map, cmap=cmap, interpolation='nearest', extent=extent, alpha=0.5)
    plot_bbox(obstacle_map.grid.bbox, 'C0--', lw=0.2, dashes=[20, 20])


def show_distance_map(distance_map):
    extent = bbox_to_extent(distance_map.grid.bbox)
    pl.imshow(distance_map.map, cmap=pl.cm.gray, interpolation='nearest', extent=extent)
    plot_bbox(distance_map.grid.bbox, 'C0--', lw=0.2, dashes=[20, 20])


def bbox_to_extent(bbox):
    return [bbox[0], bbox[0] + bbox[2], bbox[1] + bbox[3], bbox[1]]


def plot_bbox(bbox, *args, **kwargs):
    pl.plot([bbox[0], bbox[0] + bbox[2], bbox[0] + bbox[2], bbox[0], bbox[0]],
            [bbox[1], bbox[1], bbox[1] + bbox[3], bbox[1] + bbox[3], bbox[1]],
            *args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import numpy as np

from .point import Point
from .pose import Pose


@dataclass(slots=True, kw_only=True)
class Spline:
    start: Point
    control1: Point
    control2: Point
    end: Point

    a: float = 0
    b: float = 0
    c: float = 0
    d: float = 0
    e: float = 0
    f: float = 0
    g: float = 0
    h: float = 0

    m: float = 0
    n: float = 0
    o: float = 0
    p: float = 0
    q: float = 0
    r: float = 0

    def __post_init__(self) -> None:
        self.a = self.start.x
        self.e = self.start.y
        self.b = self.control1.x
        self.f = self.control1.y
        self.c = self.control2.x
        self.g = self.control2.y
        self.d = self.end.x
        self.h = self.end.y

        self.m = self.d - 3 * self.c + 3 * self.b - self.a
        self.n = self.c - 2 * self.b + self.a
        self.o = self.b - self.a
        self.p = self.h - 3 * self.g + 3 * self.f - self.e
        self.q = self.g - 2 * self.f + self.e
        self.r = self.f - self.e

    def __repr__(self) -> str:
        return f'{type(self).__qualname__}(' + ' ~ '.join([
            f'{self.start.x:.3f},{self.start.y:.3f}',
            f'{self.control1.x:.3f},{self.control1.y:.3f}',
            f'{self.control2.x:.3f},{self.control2.y:.3f}',
            f'{self.end.x:.3f},{self.end.y:.3f}',
        ]) + ')'

    @staticmethod
    def from_poses(start: Pose, end: Pose, *, control_dist: Optional[float] = None, backward: bool = False) -> Spline:
        distance = 0.5 * (start.distance(end) if control_dist is None else control_dist) * (-1 if backward else 1)
        return Spline(
            start=start.point,
            control1=start.point.polar(distance, start.yaw),
            control2=end.point.polar(-distance, end.yaw),
            end=end.point,
        )

    @staticmethod
    def from_points(start: Point, end: Point) -> Spline:
        return Spline(
            start=start,
            control1=start.interpolate(end, 0.5),
            control2=start.interpolate(end, 0.5),
            end=end,
        )

    def x(self, t: float) -> float:
        return t**3 * self.d + 3 * t**2 * (1 - t) * self.c + 3 * t * (1 - t)**2 * self.b + (1 - t)**3 * self.a

    def y(self, t: float) -> float:
        return t**3 * self.h + 3 * t**2 * (1 - t) * self.g + 3 * t * (1 - t)**2 * self.f + (1 - t)**3 * self.e

    def gx(self, t: float) -> float:
        return 3 * (self.m * t**2 + 2 * self.n * t + self.o)

    def ggx(self, t: float) -> float:
        return 6 * (self.m * t + self.n)

    def gy(self, t: float) -> float:
        return 3 * (self.p * t**2 + 2 * self.q * t + self.r)

    def ggy(self, t: float) -> float:
        return 6 * (self.p * t + self.q)

    def yaw(self, t: float) -> float:
        return np.arctan2(self.gy(t), self.gx(t))

    def pose(self, t: float) -> Pose:
        return Pose(x=self.x(t), y=self.y(t), yaw=self.yaw(t))

    def curvature(self, t: float) -> float:
        x_ = self.gx(t)
        y_ = self.gy(t)
        x__ = self.ggx(t)
        y__ = self.ggy(t)
        return (x_ * y__ - y_ * x__) / (x_**2 + y_**2)**(3/2)

    def max_curvature(self, t_min: float = 0.0, t_max: float = 1.0) -> float:
        poly = [
            (1296 * self.m * self.p ** 2 + 1296 * self.m ** 3) * self.q - 1296 * self.n * self.p ** 3 - 1296 * self.m ** 2 *
            self.n * self.p, (1620 * self.m * self.p ** 2 + 1620 * self.m ** 3) * self.r + 3240 * self.m * self.p * self.q
            ** 2 + (3240 * self.m ** 2 * self.n - 3240 * self.n * self.p ** 2) * self.q - 1620 * self.o * self.p ** 3 +
            ((-1620 * self.m ** 2 * self.o) - 3240 * self.m * self.n ** 2) * self.p,
            (5184 * self.m * self.p * self.q + 1296 * self.n * self.p ** 2 + 6480 * self.m ** 2 * self.n) * self.r + 1296 *
            self.m * self.q ** 3 - 1296 * self.n * self.p * self.q ** 2 +
            ((-6480 * self.o * self.p ** 2) - 1296 * self.m ** 2 * self.o + 1296 * self.m * self.n ** 2) * self.q +
            ((-5184 * self.m * self.n * self.o) - 1296 * self.n ** 3) * self.p, 1296 * self.m * self.p * self.r ** 2 +
            (1944 * self.m * self.q ** 2 + 6480 * self.n * self.p * self.q - 1296 * self.o * self.p ** 2 + 1296 * self.m **
             2 * self.o + 8424 * self.m * self.n ** 2) * self.r - 8424 * self.o * self.p * self.q ** 2 - 6480 * self.m *
            self.n * self.o * self.q + ((-1296 * self.m * self.o ** 2) - 1944 * self.n ** 2 * self.o) * self.p, 2592 *
            self.n * self.p * self.r ** 2 +
            (3888 * self.n * self.q ** 2 - 2592 * self.o * self.p * self.q + 2592 * self.m * self.n * self.o + 3888 * self.
             n ** 3) * self.r - 3888 * self.o * self.q ** 3 +
            ((-2592 * self.m * self.o ** 2) - 3888 * self.n ** 2 * self.o) * self.q, -324 * self.m * self.r ** 3 +
            (1944 * self.n * self.q + 324 * self.o * self.p) * self.r ** 2 +
            ((-1944 * self.o * self.q ** 2) - 324 * self.m * self.o ** 2 + 1944 * self.n ** 2 * self.o) * self.r - 1944 *
            self.n * self.o ** 2 * self.q + 324 * self.o ** 3 * self.p, ]
        roots = np.roots(poly)
        t = np.array([t0 for t0 in roots if np.isreal(t0) and t_min < t0 < t_max] + [t_min, t_max])

        k = np.real(self.curvature(t))
        idx = np.argmax(np.abs(k))
        return k[idx]

    def closest_point(self, x: float, y: float, t_min: float = 0.0, t_max: float = 1.0) -> float:
        poly = [
            6 * self.h**2 + ((-36 * self.g) + 36 * self.f - 12 * self.e) * self.h + 54 * self.g**2 + (36 * self.e - 108 * self.f) * self.g + 54 * self.f**2 - 36 * self.e * self.f + 6 * self.e**2 +
            6 * self.d**2 + ((-36 * self.c) + 36 * self.b - 12 * self.a) * self.d + 54 * self.c**2 +
            (36 * self.a - 108 * self.b) * self.c + 54 * self.b**2 - 36 * self.a * self.b + 6 * self.a**2,
            (30 * self.g - 60 * self.f + 30 * self.e) * self.h - 90 * self.g**2 + (270 * self.f - 120 * self.e) * self.g - 180 * self.f**2 + 150 * self.e * self.f - 30 * self.e**2 +
            (30 * self.c - 60 * self.b + 30 * self.a) * self.d - 90 * self.c**2 + (270 * self.b -
                                                                                   120 * self.a) * self.c - 180 * self.b**2 + 150 * self.a * self.b - 30 * self.a**2,
            (24 * self.f - 24 * self.e) * self.h + 36 * self.g**2 + (144 * self.e - 216 * self.f) * self.g + 216 * self.f**2 - 240 * self.e * self.f + 60 * self.e**2 +
            (24 * self.b - 24 * self.a) * self.d + 36 * self.c**2 + (144 * self.a - 216 *
                                                                     self.b) * self.c + 216 * self.b**2 - 240 * self.a * self.b + 60 * self.a**2,
            ((-6 * self.h) + 18 * self.g - 18 * self.f + 6 * self.e) * y + ((-6 * self.d) + 18 * self.c - 18 * self.b + 6 * self.a) * x + 6 * self.e * self.h + (54 * self.f - 72 * self.e) *
            self.g - 108 * self.f**2 + 180 * self.e * self.f - 60 * self.e**2 + 6 * self.a * self.d +
            (54 * self.b - 72 * self.a) * self.c - 108 * self.b**2 + 180 * self.a * self.b - 60 * self.a**2,
            ((-12 * self.g) + 24 * self.f - 12 * self.e) * y + ((-12 * self.c) + 24 * self.b - 12 * self.a) * x + 12 * self.e * self.g + 18 *
            self.f**2 - 60 * self.e * self.f + 30 * self.e**2 + 12 * self.a *
            self.c + 18 * self.b**2 - 60 * self.a * self.b + 30 * self.a**2,
            (6 * self.e - 6 * self.f) * y + (6 * self.a - 6 * self.b) * x + 6 *
            self.e * self.f - 6 * self.e**2 + 6 * self.a * self.b - 6 * self.a**2,
        ]

        roots = np.roots(poly)
        t = np.array([t0 for t0 in roots if np.isreal(t0) and t_min < t0 < t_max] + [t_min, t_max])

        sqr_d = (self.x(t) - x)**2 + (self.y(t) - y)**2

        return np.real(t[np.argmin(sqr_d)])

    def turning_points(self, t_min: float = 0.0, t_max: float = 1.0) -> np.ndarray:
        inner = self.m**2 * self.r**2 + ((4 * self.n**2 - 2 * self.m * self.o) * self.p - 4 * self.m * self.n * self.q) * \
            self.r + 4 * self.m * self.o * self.q**2 - 4 * self.n * self.o * self.p * self.q + self.o**2 * self.p**2
        if inner < 0:
            return np.array([])

        denominator = 2 * self.m * self.q - 2 * self.n * self.p
        if abs(denominator) < 1e-16:
            return np.array([])

        t = [
            +(np.sqrt(inner) - self.m * self.r + self.o * self.p) / denominator,
            -(np.sqrt(inner) + self.m * self.r - self.o * self.p) / denominator,
        ]
        return np.array([t_ for t_ in t if t_min <= t_ <= t_max])

    def estimated_length(self, steps: int = 10) -> float:
        dx = np.diff([self.x(t) for t in np.linspace(0, 1, steps)])
        dy = np.diff([self.y(t) for t in np.linspace(0, 1, steps)])
        return np.sum(np.sqrt(dx**2 + dy**2))



from dataclasses import dataclass
from typing import Any

from . import rosys


@dataclass(slots=True, kw_only=True)
class Config:
    ui_update_interval: float = 0.1
    simulation_speed: float = 1.0
    needs_backup: bool = False

    def __post_init__(self) -> None:
        rosys.persistence.register(self)

    def backup(self) -> dict[str, Any]:
        return {'simulation_speed': self.simulation_speed}

    def restore(self, data: dict[str, Any]) -> None:
        self.simulation_speed = data.get('simulation_speed', 1.0)



import logging
from typing import Callable, Coroutine, Optional

from .. import rosys
from ..analysis import track
from ..driving import Steerer
from ..event import Event
from .automation import Automation


class Automator:
    """An automator allows running automations, i.e. coroutines that can be paused and resumed.

    See [Click-and-drive](https://rosys.io/examples/click-and-drive/) for a simple example of an automation.

    _steerer_: If provided, manually steering the robot will pause a currently running automation.

    _default_automation_: If provided, it allows the automator to start a new automation without passing an automation
    (e.g. via an "Play"-button like offered by the [automation controls](https://rosys.io/reference/rosys/automation/#rosys.automation.automation_controls)). 
    The passed function should return a new coroutine on every call (see [Play-pause-stop](https://rosys.io/examples/play-pause-stop/) example).

    _on_interrupt_: Optional callback that will be called when an automation pauses or stops.
    """

    def __init__(self,
                 steerer: Optional[Steerer], *,
                 default_automation: Optional[Callable] = None,
                 on_interrupt: Optional[Callable] = None) -> None:
        self.AUTOMATION_STARTED = Event()
        """an automation has been started"""

        self.AUTOMATION_PAUSED = Event()
        """an automation has been paused (string argument: description of the cause)"""

        self.AUTOMATION_RESUMED = Event()
        """an automation has been resumed"""

        self.AUTOMATION_STOPPED = Event()
        """an automation has been stopped (string argument: description of the cause)"""

        self.AUTOMATION_FAILED = Event()
        """an automation has failed to complete (string argument: description of the cause)"""

        self.AUTOMATION_COMPLETED = Event()
        """an automation has been completed"""

        self.log = logging.getLogger('rosys.automator')

        self.default_automation = default_automation

        self.enabled: bool = True
        self.automation: Optional[Automation] = None

        if steerer:
            steerer.STEERING_STARTED.register(lambda: self.pause(because='steering started'))

        if on_interrupt:
            self.AUTOMATION_PAUSED.register(lambda _: rosys.background_tasks.create(on_interrupt()))
            self.AUTOMATION_STOPPED.register(lambda _: rosys.background_tasks.create(on_interrupt()))

        rosys.on_shutdown(lambda: self.stop(because='automator is shutting down'))

    @property
    def is_stopped(self) -> bool:
        return self.automation is None or self.automation.is_stopped

    @property
    def is_running(self) -> bool:
        return self.automation is not None and self.automation.is_running

    @property
    def is_paused(self) -> bool:
        return self.automation is not None and self.automation.is_paused

    def start(self, coro: Optional[Coroutine] = None, *, paused: bool = False) -> None:
        """Starts a new automation.

        You can pass any coroutine.
        The automator will make sure it can be paused, resumed and stopped.
        """
        if coro is None:
            self.start(self.default_automation(), paused=paused)
            return
        if not self.enabled:
            coro.close()
            return
        self.stop(because='new automation starts')
        self.automation = Automation(coro, self._handle_exception, on_complete=self._on_complete)
        rosys.background_tasks.create(self.automation.__await__(), name='automation')
        self.AUTOMATION_STARTED.emit()
        rosys.notify('automation started')
        if paused:
            self.automation.pause()

    def pause(self, because: str) -> None:
        """Pauses the current automation.

        You need to provide a cause which will be used as notification message.
        """
        if self.is_running:
            self.automation.pause()
            self.AUTOMATION_PAUSED.emit(because)
            rosys.notify(f'automation paused because {because}')

    def resume(self) -> None:
        """Resumes the current automation."""
        if not self.enabled:
            return
        if self.is_paused:
            self.automation.resume()
            self.AUTOMATION_RESUMED.emit()
            rosys.notify('automation resumed')

    def stop(self, because: str) -> None:
        """Stops the current automation.

        You need to provide a cause which will be used as notification message.
        """
        if not self.is_stopped:
            self.automation.stop()
            self.AUTOMATION_STOPPED.emit(because)
            track.reset()
            rosys.notify(f'automation stopped because {because}')

    def enable(self) -> None:
        """Enables the automator.

        It is enabled by default.
        It can be disabled by calling `disable()`.
        """
        self.enabled = True

    def disable(self, because: str) -> None:
        """Disables the automator.

        No automations can be started while the automator is disabled.
        If an automation is running or paused it will be stopped.
        You need to provide a cause which will be used as notification message.
        """
        self.stop(because)
        self.enabled = False

    def _handle_exception(self, e: Exception) -> None:
        self.stop(because='an exception occurred in an automation')
        self.AUTOMATION_FAILED.emit(str(e))
        rosys.notify('automation failed', 'negative')
        if rosys.is_test:
            self.log.exception('automation failed', e)

    def _on_complete(self) -> None:
        self.AUTOMATION_COMPLETED.emit()
        rosys.notify('automation completed', 'positive')



from __future__ import annotations

import asyncio
import inspect
import logging
from dataclasses import dataclass
from typing import Awaitable, Callable

from nicegui import globals as nicegui_globals

from rosys import background_tasks

from .helpers import invoke

startup_coroutines: list[Awaitable] = []
tasks: list[asyncio.Task] = []
log = logging.getLogger('rosys.event')
events: list[Event] = []


@dataclass(slots=True, kw_only=True)
class EventListener:
    callback: Callable
    filepath: str
    line: int


class Event:

    def __init__(self) -> None:
        self.listeners: list[EventListener] = list()
        events.append(self)

    def register(self, callback: Callable) -> Event:
        if not callable(callback):
            raise Exception('non-callable callback')
        if any(l.callback == callback for l in self.listeners):
            return  # NOTE: don't add duplicate listeners
        frame = inspect.currentframe().f_back
        self.listeners.append(EventListener(callback=callback, filepath=frame.f_code.co_filename, line=frame.f_lineno))
        return self

    def register_ui(self, callback: Callable) -> Event:
        self.register(callback)
        client = nicegui_globals.get_client()
        if not client.shared:
            async def register_disconnect():
                try:
                    await client.connected(timeout=10.0)
                    client.on_disconnect(lambda: self.unregister(callback))
                except TimeoutError:
                    log.warning(f'could not register disconnect for {callback=}')
                    self.unregister(callback)
            background_tasks.create(register_disconnect())
        return self

    def unregister(self, callback: Callable) -> None:
        self.listeners[:] = [l for l in self.listeners if l.callback != callback]

    async def call(self, *args) -> None:
        """Fires event and waits async until all registered listeners are completed"""
        for listener in self.listeners:
            try:
                await invoke(listener.callback, *args)
            except Exception:
                log.exception(f'could not call {listener=}')

    def emit(self, *args) -> None:
        """Fires event without waiting for the result."""
        for listener in self.listeners:
            try:
                result = invoke(listener.callback, *args)
                if isinstance(result, Awaitable):
                    if nicegui_globals.loop.is_running():
                        name = f'{listener.filepath}:{listener.line}'
                        tasks.append(background_tasks.create(result, name=name))
                    else:
                        startup_coroutines.append(result)
            except Exception:
                log.exception(f'could not emit {listener=}')


def reset() -> None:
    for event in events:
        event.listeners.clear()
    for task in tasks:
        task.cancel()
    tasks.clear()



#!/usr/bin/env python3
import time

import numpy as np
import pylab as pl
from nicegui import ui

from .binary_renderer import BinaryRenderer

renderer = BinaryRenderer((30, 40), fill_value=True)

area_outline = np.array([[-5, 5], [35, 3], [40, 25], [5, 30]])
t = time.time()
renderer.polygon(area_outline, False)
ui.label(f'area: {(time.time() - t) * 1000:.3f} ms')

x, y, r = 10, 15, 5
t = time.time()
renderer.circle(x, y, r)
ui.label(f'circle: {(time.time() - t) * 1000:.3f} ms')

polygon_outline = np.array([[20.5, 10], [30, 15], [25, 25]])
t = time.time()
renderer.polygon(polygon_outline)
ui.label(f'polygon: {(time.time() - t) * 1000:.3f} ms')

with ui.plot():
    pl.imshow(renderer.map, cmap=pl.cm.gray, interpolation='nearest')
    pl.xlim(0, renderer.map.shape[1] - 1)
    pl.ylim(renderer.map.shape[0] - 1, 0)
    pl.gca().add_patch(pl.matplotlib.patches.Circle((x, y), r, color='none', ec='C0', lw=2))
    pl.fill(polygon_outline[:, 0], polygon_outline[:, 1], color='none', ec='C0', lw=2)
    pl.fill(area_outline[:, 0], area_outline[:, 1], color='none', ec='C2', lw=2)

ui.run()



import numpy as np
from nicegui.elements.scene_objects import Extrusion, Group

from .area import Area
from .path_planner import PathPlanner


class AreaObject(Group):

    def __init__(self, path_planner: PathPlanner) -> None:
        super().__init__()

        self.update(path_planner.areas)
        path_planner.AREAS_CHANGED.register_ui(self.update)

    def update(self, areas: dict[str, Area]) -> None:
        [obj.delete() for obj in list(self.scene.objects.values()) if (obj.name or '').startswith('area_')]
        for area in areas.values():
            if len(area.outline) == 1:
                outline = [[area.outline[0].x + 0.05 * np.cos(phi), area.outline[0].y + 0.05 * np.sin(phi)]
                           for phi in np.linspace(0, 2 * np.pi, 16, endpoint=False)]
            else:
                outline = [[point.x, point.y] for point in area.outline]
            with self.scene:
                Extrusion(outline, 0.1, wireframe=True).with_name(f'area_{area.id}').material(area.color)



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from typing import Optional

import numpy as np
from rosys import persistence

from ..geometry import Rotation
from .calibration import Calibration, Extrinsics, Intrinsics
from .image import Image, ImageSize


@dataclass(slots=True, kw_only=True)
class Camera(abc.ABC):
    id: str
    calibration: Optional[Calibration] = None
    images: list[Image] = field(default_factory=list, metadata=persistence.exclude)
    focal_length: Optional[float] = None

    @property
    def captured_images(self) -> list[Image]:
        return [i for i in self.images if i.data]

    @property
    def latest_captured_image(self) -> Optional[Image]:
        images = self.captured_images
        if images:
            return images[-1]

    def get_recent_images(self, current_time: float, timespan: float = 10.0) -> list[Image]:
        return [i for i in self.captured_images if i.time > current_time - timespan]

    def set_perfect_calibration(
        self,
        *,
        width=800, height=600,
        x: float = 0.0, y: float = 0.0, z: float = 1.0,
        roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0,
    ) -> None:
        self.calibration = Calibration(
            intrinsics=Camera.create_intrinsics(width, height),
            extrinsics=Extrinsics(rotation=Rotation.from_euler(roll, pitch, yaw), translation=[x, y, z]),
        )

    @staticmethod
    def create_intrinsics(width: int = 800, height: int = 600) -> Intrinsics:
        c = 570
        size = ImageSize(width=width, height=height)
        K = [[c, 0, size.width / 2], [0, c, size.height / 2], [0, 0, 1]]
        D = [0, 0, 0, 0, 0]
        rotation = Rotation.zero()
        return Intrinsics(matrix=K, distortion=D, rotation=rotation, size=size)



from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from ..geometry import Pose


@dataclass
class FastSpline:
    start_x: float
    start_y: float
    start_yaw: float
    end_x: float
    end_y: float
    end_yaw: float
    backward: bool

    def __post_init__(self) -> None:
        distance = -0.5 if self.backward else 0.5
        distance *= np.sqrt((self.end_x - self.start_x)**2 + (self.end_y - self.start_y)**2)
        self.a = self.start_x
        self.e = self.start_y
        self.b = self.start_x + distance * np.cos(self.start_yaw)
        self.f = self.start_y + distance * np.sin(self.start_yaw)
        self.c = self.end_x - distance * np.cos(self.end_yaw)
        self.g = self.end_y - distance * np.sin(self.end_yaw)
        self.d = self.end_x
        self.h = self.end_y
        self.m = self.d - 3 * self.c + 3 * self.b - self.a
        self.n = self.c - 2 * self.b + self.a
        self.o = self.b - self.a
        self.p = self.h - 3 * self.g + 3 * self.f - self.e
        self.q = self.g - 2 * self.f + self.e
        self.r = self.f - self.e

    @staticmethod
    def from_poses(start: Pose, end: Pose, *, backward: bool = False) -> FastSpline:
        return FastSpline(start.x, start.y, start.yaw, end.x, end.y, end.yaw, backward)

    def x(self, t: float) -> float:
        return t**3 * self.d + 3 * t**2 * (1 - t) * self.c + 3 * t * (1 - t)**2 * self.b + (1 - t)**3 * self.a

    def y(self, t: float) -> float:
        return t**3 * self.h + 3 * t**2 * (1 - t) * self.g + 3 * t * (1 - t)**2 * self.f + (1 - t)**3 * self.e

    def gx(self, t: float) -> float:
        return 3 * (self.m * t**2 + 2 * self.n * t + self.o)

    def gy(self, t: float) -> float:
        return 3 * (self.p * t**2 + 2 * self.q * t + self.r)

    def yaw(self, t: float) -> float:
        return np.arctan2(self.gy(t), self.gx(t))



from __future__ import annotations

import io
import logging
import os
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Optional, Protocol

import humanize
from PIL import Image, ImageDraw, ImageFont

from .. import rosys

STORAGE_PATH = os.path.expanduser('~/.rosys/timelapse')
FONT = f'{os.path.dirname(__file__)}/assets/RobotoMono-Medium.ttf'
IMAGE_FONT = ImageFont.truetype(FONT, 12)
BIG_COVER_FONT = ImageFont.truetype(FONT, 100)
SMALL_COVER_FONT = ImageFont.truetype(FONT, 60)


class RosysImage(Protocol):
    camera_id: str
    time: float
    data: Optional[bytes] = None


class TimelapseRecorder:

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.timelapse_recorder')

        os.makedirs(STORAGE_PATH + '/videos', exist_ok=True)

    async def save(self, image: RosysImage) -> None:
        await rosys.run.cpu_bound(save_image, image, STORAGE_PATH)

    async def compress_video(self) -> None:
        jpgs = sorted(glob(f'{STORAGE_PATH}/*.jpg'))
        if len(jpgs) < 20:
            self.log.info(f'very few images ({len(jpgs)}); not creating video')
            return
        start_unix = int(Path(jpgs[0]).stem[:-4])
        start = datetime.fromtimestamp(start_unix)
        end = datetime.fromtimestamp(int(Path(jpgs[-1]).stem[:-4]))
        self.log.info(f'creating video from {start} to {end}')
        duration = humanize.naturaldelta(end-start)
        await self.create_info(start.strftime('%d.%m.%Y %H:%M:%S'), duration, time=start_unix - 1)
        id = start.strftime('%Y%m%d_%H-%M-%S_' + duration.replace(' ', '_'))
        target_dir = STORAGE_PATH + '/' + id
        os.makedirs(target_dir, exist_ok=True)
        os.makedirs(STORAGE_PATH + '/videos', exist_ok=True)
        self.log.info(await rosys.run.sh(f'mv {STORAGE_PATH}/*.jpg {target_dir}', shell=True))
        # it seems that niceness of subprocess is relative to own niceness, but we want an absolute niceness
        absolute_niceness = 10 - os.nice(0)
        cmd = f'nice -n {absolute_niceness} ffmpeg -hide_banner -threads 1 -r 10 -pattern_type glob -i "{target_dir}/*.jpg" -s 1600x1200 -vcodec libx264 -crf 18 -preset slow -pix_fmt yuv420p -y {target_dir}/{id}.mp4; mv {target_dir}/*mp4 {STORAGE_PATH}/videos; rm -r {target_dir};'
        self.log.info(f'starting {cmd}')
        rosys.background_tasks.create(rosys.run.sh(cmd, timeout=None, shell=True), name='timelapse ffmpeg')

    def clear_jpegs(self) -> None:
        files = glob(f'{STORAGE_PATH}/**/*.jpg', recursive=True)
        for f in files:
            os.remove(f)

    async def create_info(self, title: str, subtitle: str, frames: int = 20, time: Optional[int] = None) -> None:
        prefix = f'{STORAGE_PATH}/{int(rosys.time() if time is None else time)}'
        await rosys.run.cpu_bound(save_info, title, subtitle, prefix, frames)


def save_image(image: RosysImage, path: str) -> None:
    img = Image.open(io.BytesIO(image.data))
    draw = ImageDraw.Draw(img)
    x = img.width - 300
    y = img.height - 18
    text = f'{datetime.fromtimestamp(image.time).strftime("%Y-%m-%d %H:%M:%S")}, cam {image.camera_id}'
    # shadow
    draw.text((x - 1, y - 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x + 1, y - 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x - 1, y + 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x + 1, y + 1), text, font=IMAGE_FONT, fill=(0, 0, 0))
    draw.text((x, y), text, font=IMAGE_FONT, fill=(255, 255, 255))
    filepath = path + f'/{int(image.time)}0000.jpg'
    img.save(filepath, 'JPEG')


def save_info(title: str, subtitle: str, prefix: str, frames: int) -> None:
    for i in range(frames):
        img = Image.new('RGB', (1600, 1200), 'black')
        draw = ImageDraw.Draw(img)
        font = BIG_COVER_FONT if len(subtitle) < 25 and len(title) < 25 else SMALL_COVER_FONT
        w, h = draw.textsize(title, font=font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 - 200), title, font=font, fill='white')
        w, h = draw.textsize(subtitle, font=font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 + 100), subtitle, font=font, fill='white')
        filepath = f'{prefix}{i:04}.jpg'
        img.save(filepath)



import gc
import logging
import os
import tracemalloc

import psutil
from fastapi import Request
from psutil._common import bytes2human
from starlette.middleware.base import BaseHTTPMiddleware

import rosys

log = logging.getLogger('rosys.analysis.memory')


class MemoryMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next):
        mem_before = get_process_memory()
        response = await call_next(request)
        if response.headers.get('x-nicegui-content') == 'page':
            self.mem_before = get_process_memory()
            mem_after = get_process_memory()
            msg = f'GET {request.get("path")} increased memory by {bytes2human(mem_after - mem_before)} and is now {bytes2human(mem_after)}'
            log.info(msg)
        return response


def get_process_memory() -> int:
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


def get_humanreadable_process_memory() -> int:
    return bytes2human(get_process_memory())


def compare_tracemalloc_snapshots(snapshot, prev_snapshot):
    stats = snapshot.compare_to(prev_snapshot, 'traceback')
    for stat in stats[:10]:
        trace = '\n'.join(stat.traceback.format())
        usage = f'{bytes2human(stat.size_diff)} new, {bytes2human(stat.size)} total; {stat.count_diff} new memory blocks, {stat.count} total'
        log.info(trace + '\n' + usage)


def observe_memory_growth(with_tracemalloc: bool = False) -> None:
    log.info('Observing memory growth')
    prev_memory: int = 0
    prev_snapshot: tracemalloc.Snapshot = None
    if with_tracemalloc:
        tracemalloc.start(10)

    async def stats() -> None:
        nonlocal prev_memory
        nonlocal prev_snapshot
        gc.collect()
        growth = rosys.analysis.memory.get_process_memory() - prev_memory
        # log.info('==============')
        log.info(
            f'memory growth: {bytes2human(growth)}, '
            f'now it\'s {rosys.analysis.memory.get_humanreadable_process_memory()}'
        )
        # log.info('==============')
        prev_memory = rosys.analysis.memory.get_process_memory()
        if with_tracemalloc:
            snapshot = tracemalloc.take_snapshot()
            if growth > 4 * 1e-6 and prev_snapshot is not None:
                await rosys.run.cpu_bound(compare_tracemalloc_snapshots, snapshot, prev_snapshot)
            prev_snapshot = snapshot

    rosys.on_repeat(stats, 60.0)



import json
import logging
import os
import sys
from dataclasses import fields
from typing import Any, Callable, Optional, Protocol, TypeVar

import numpy as np
from dataclasses_json import Exclude, config
from dataclasses_json.core import _asdict, _decode_dataclass
from nicegui import app, ui
from nicegui.events import UploadEventArguments
from starlette.responses import FileResponse

from .helpers import invoke
from .run import awaitable

exclude = config(exclude=Exclude.ALWAYS)


def to_dict(obj: Any) -> dict[str, Any]:
    return _asdict(obj, False)


def from_dict(cls: type, d: dict[str, Any]) -> Any:
    return _decode_dataclass(cls, d, False)


T = TypeVar('T')


def replace_dict(old_dict: dict[str, T], cls: type, new_dict: dict[str, T]) -> None:
    """Replace content of `old_dict` with keys and values from `new_dict`."""
    old_dict.clear()
    old_dict.update({key: from_dict(cls, value) for key, value in new_dict.items()})


def replace_list(old_list: list[T], cls: type, new_list: list[T]) -> None:
    """Replace content of `old_list` with items from `new_list`."""
    old_list.clear()
    old_list.extend(from_dict(cls, value) for value in new_list)


def replace_dataclass(old_dataclass: Any, new_dict: dict[str, Any]) -> None:
    """Replace content of `old_dataclass` with content from `new_dict`."""
    for field in fields(old_dataclass):
        setattr(old_dataclass, field.name, new_dict.get(field.name))


backup_path = os.path.expanduser('~/.rosys')
log = logging.getLogger('rosys.persistence')

is_test = 'pytest' in sys.modules


class PersistentModule(Protocol):
    needs_backup: bool

    def backup(self) -> dict[str, Any]:
        ...

    def restore(self, data: dict[str, Any]) -> None:
        ...


modules: dict[str, PersistentModule] = {}


def register(module: PersistentModule) -> None:
    if not is_test:
        modules[module.__module__] = module


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.floating):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


@awaitable
def backup(force: bool = False) -> None:
    for name, module in modules.items():
        if not module.needs_backup and not force:
            continue
        if not os.path.exists(backup_path):
            os.makedirs(backup_path)
        filepath = f'{backup_path}/{name}.json'
        with open(filepath, 'w') as f:
            json.dump(module.backup(), f, indent=4, cls=Encoder)
        module.needs_backup = False


def restore() -> None:
    for name, module in modules.items():
        filepath = f'{backup_path}/{name}.json'
        if not os.path.exists(filepath):
            log.warning(f'Backup file "{filepath}" not found.')
            continue
        with open(filepath) as f:
            try:
                module.restore(json.load(f))
            except Exception:
                log.exception(f'failed to restore {module}')


def export_button(title: str = 'Export', route: str = '/export', tmp_filepath: str = '/tmp/export.json') -> ui.button:
    @app.get(route)
    def get_export() -> FileResponse:
        with open(tmp_filepath, 'w') as f:
            json.dump({name: module.backup() for name, module in modules.items()}, f, indent=4)
        return FileResponse(tmp_filepath, filename='export.json')
    ui.button(title, on_click=lambda: ui.download(route[1:]))


def import_button(title: str = 'Import', after_import: Optional[Callable] = None) -> ui.button:
    async def restore_from_file(e: UploadEventArguments) -> None:
        all_data = json.load(e.content)
        assert isinstance(all_data, dict)
        for name, data in all_data.items():
            modules[name].restore(data)
        await backup(force=True)
        dialog.close()
        await invoke(after_import)
    with ui.dialog() as dialog, ui.card():
        ui.upload(on_upload=restore_from_file)
    return ui.button(title, on_click=dialog.open)



from .calibration import Calibration, Extrinsics, Intrinsics
from .camera import Camera
from .camera_objects_ import CameraObjects as camera_objects
from .camera_projector import CameraProjector
from .camera_provider import CameraProvider
from .detections import BoxDetection, Detection, Detections, PointDetection
from .detector import Detector
from .detector_hardware import DetectorHardware
from .detector_simulation import DetectorSimulation, SimulatedObject
from .image import Image, ImageSize
from .multi_camera_provider import MultiCameraProvider
from .usb_camera import UsbCamera
from .usb_camera_provider_hardware import UsbCameraProviderHardware
from .usb_camera_provider_simulation import UsbCameraProviderSimulation



import inspect
import time
from contextlib import contextmanager
from typing import Any, Awaitable, Callable, Generator

import numpy as np


async def invoke(handler: Callable, *args: Any) -> Any:
    result = handler(*args)
    if isinstance(result, Awaitable):
        result = await result
    return result


def measure(*, reset: bool = False, ms: bool = False) -> None:
    global t
    if 't' in globals() and not reset:
        dt = time.perf_counter() - t
        line = inspect.stack()[1][0].f_lineno
        output = f'{dt * 1000:7.3f} ms' if ms else f'{dt:7.3f} s'
        print(f'{inspect.stack()[1].filename}:{line}', output, flush=True)
    if reset:
        print('------------', flush=True)
    t = time.perf_counter()


def angle(yaw0: float, yaw1: float) -> float:
    return eliminate_2pi(yaw1 - yaw0)


def eliminate_pi(angle: float) -> float:
    return (angle + np.pi / 2) % np.pi - np.pi / 2


def eliminate_2pi(angle: float) -> float:
    return (angle + np.pi) % (2 * np.pi) - np.pi


def ramp(x: float, in_min: float, in_max: float, out_min: float, out_max: float, clip: bool = False) -> float:
    if clip and x < in_min:
        return out_min
    if clip and x > in_max:
        return out_max
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min


def remove_indentation(text: str) -> str:
    """Remove indentation from a multi-line string based on the indentation of the first line."""
    lines = text.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    if not lines:
        return ''
    indentation = len(lines[0]) - len(lines[0].lstrip())
    return '\n'.join(line[indentation:] for line in lines)


class ModificationContext:

    @contextmanager
    def set(self, **kwargs) -> Generator[None, None, None]:
        backup = {key: getattr(self, key) for key in kwargs.keys()}
        for key, value in kwargs.items():
            setattr(self, key, value)
        try:
            yield
        finally:
            for key, value in backup.items():
                setattr(self, key, value)



import logging
import os.path
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional

from ... import rosys


@dataclass(slots=True, kw_only=True)
class Measurement:
    time: str
    duration: float
    name: str
    details: str


color_pattern = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
warning_pattern = re.compile(r"(.*) \[WARNING\].*Executing(.*)took (.*) seconds")
task_pattern = re.compile(r".*name=['\"](.*)['\"] coro=<(.*)> .*")
coro_pattern = re.compile(r"(.*) (running at .*)")


class AsyncioMonitor:

    def __init__(self, log_filepath: str = os.path.expanduser('~/.rosys/debug.log')) -> None:
        self.log = logging.getLogger('rosys.asyncio_monitor')

        self.log_filepath = log_filepath
        self.timings: dict[str, list[Measurement]] = defaultdict(list)
        self.log_position: Optional[int] = None

        rosys.on_repeat(self.step, 10)

    async def step(self) -> None:
        if not os.path.isfile(self.log_filepath):
            return

        if self.log_position is None:  # NOTE only messages generated after startup should be analyzed
            self.log_position = os.path.getsize(self.log_filepath)
        if os.path.getsize(self.log_filepath) < self.log_position:  # NOTE detect beginning of new log file
            self.log_position = 0
        ignore_warnings = False
        with open(self.log_filepath, 'r') as f:
            f.seek(self.log_position)
            for line in f:
                line = color_pattern.sub('', line)
                if 'start garbage collection' in line:
                    ignore_warnings = True
                if 'finished garbage collection' in line:
                    ignore_warnings = False
                if not ignore_warnings:
                    message = self.parse_async_warning(line)
                    if message:
                        self.timings[message.name].append(message)
            self.log_position = f.tell()

    def parse_async_warning(self, msg: str) -> Optional[Measurement]:
        if 'rosys/actors/asyncio_monitor.py' in msg:
            return  # NOTE we ignore our own messages
        match_warning = warning_pattern.match(msg)
        if match_warning is None:
            return
        time = match_warning.group(1)
        duration = float(match_warning.group(3))
        description = match_warning.group(2)
        if 'TimerHandle' in description:
            name = 'TimerHandle'
            details = description
        elif '<Handle' in description:
            name = 'Handle'
            details = description
        else:
            match_task = task_pattern.match(description)
            if match_task is None:
                self.log.warning(f'could not parse task "{description}"')
                return
            name = match_task.group(1)
            details = match_task.group(2)
            if name.startswith('Task-'):
                match_coro = coro_pattern.match(details)
                if match_coro is not None:
                    name = match_coro.group(1)
                    details = match_coro.group(2)
        return Measurement(time=time, name=name, duration=duration, details=details)



import asyncio
import logging
from dataclasses import fields, is_dataclass
from typing import Any, Callable, Optional

import numpy as np
import pytest

from .. import rosys, run
from ..automation import Automator
from ..driving import Driver, Odometer
from ..geometry import Point, Point3d

log = logging.getLogger(__name__)

automator: Optional[Automator] = None
driver: Optional[Driver] = None
odometer: Optional[Odometer] = None


async def forward(seconds: Optional[float] = None,
                  *,
                  until: Optional[int | float | Callable] = None,
                  x: Optional[float] = None,
                  y: Optional[float] = None,
                  tolerance: float = 0.1,
                  dt: float = 0.01,
                  timeout: float = 100):
    start_time = rosys.time()
    if seconds is not None:
        msg = f'forwarding {seconds=}'
        def condition(): return rosys.time() >= start_time + seconds
        timeout = max(timeout, seconds)
    elif isinstance(until, int) or isinstance(until, float):
        msg = f'forwarding {until=}'
        def condition(): return rosys.time() >= until
        timeout = max(timeout, until - start_time)
    elif isinstance(until, Callable):
        msg = f'forwarding {until=}'
        condition = until
    elif x is not None and y is not None:
        assert odometer is not None
        msg = f'forwarding to {x=} and {y=}'
        def condition(): return (odometer.prediction.x - x)**2 + (odometer.prediction.y - y)**2 < tolerance**2
    elif x is not None:
        assert odometer is not None
        msg = f'forwarding to {x=}'
        def condition(): return abs(odometer.prediction.x - x) < tolerance
    elif y is not None:
        assert odometer is not None
        msg = f'forwarding to {y=}'
        def condition(): return abs(odometer.prediction.y - y) < tolerance
    else:
        raise Exception('invalid arguments')

    log.info(f'\033[94m{msg}\033[0m')
    while not condition():
        if rosys.time() > start_time + timeout:
            raise TimeoutError(f'condition took more than {timeout} s')
        if not run.running_cpu_bound_processes:
            rosys.set_time(rosys.time() + dt)
            await asyncio.sleep(0)
        else:
            await asyncio.sleep(0.01)
        if rosys._exception is not None:
            raise RuntimeError(f'error while forwarding time {dt} s') from rosys._exception


def assert_pose(x: float, y: float, *, deg: float = None, position_tolerance: float = 0.1, deg_tolerance: float = 1.0) -> None:
    assert odometer is not None
    assert odometer.prediction.x == pytest.approx(x, abs=position_tolerance)
    assert odometer.prediction.y == pytest.approx(y, abs=position_tolerance)
    if deg is not None:
        assert np.rad2deg(odometer.prediction.yaw) == pytest.approx(deg, abs=deg_tolerance)


def assert_point(actual: Point | Point3d, expected: Point | Point3d, tolerance=0.1) -> None:
    assert type(actual) == type(expected)
    assert actual.x == pytest.approx(expected.x, abs=tolerance)
    assert actual.y == pytest.approx(expected.y, abs=tolerance)
    if type(actual) is Point3d:
        assert actual.z == pytest.approx(expected.z, abs=tolerance)


async def automate_drive_to(x: float, y: float) -> None:
    assert automator is not None
    assert driver is not None
    automator.start(driver.drive_to(Point(x=x, y=y)))


def approx(o1: Any, o2: Any, *, rel: Optional[float] = None, abs: Optional[float] = None, nan_ok: bool = False) -> None:
    if hasattr(o1, '__dict__'):
        assert sorted(o1.__dict__) == sorted(o2.__dict__)
        for key, value in o1.__dict__.items():
            approx(value, getattr(o2, key), rel=rel, abs=abs, nan_ok=nan_ok)
    elif is_dataclass(o1):
        o1_keys = [f.name for f in fields(o1)]
        o2_keys = [f.name for f in fields(o2)]
        assert sorted(o1_keys) == sorted(o2_keys)
        for key in o1_keys:
            approx(getattr(o1, key), getattr(o2, key), rel=rel, abs=abs, nan_ok=nan_ok)
    elif isinstance(o1, list):
        for i1, i2 in zip(o1, o2, strict=True):
            approx(i1, i2, rel=rel, abs=abs, nan_ok=nan_ok)
    elif isinstance(o1, dict):
        for k1, k2 in zip(sorted(o1), sorted(o2), strict=True):
            approx(o1[k1], o2[k2], rel=rel, abs=abs, nan_ok=nan_ok)
    else:
        assert o1 == pytest.approx(o2, rel=rel, abs=abs, nan_ok=nan_ok)



#!/usr/bin/env python3
import logging

import serial
import socketio
import uvicorn
from hardware.robot_brain import augment, check

sio = socketio.AsyncServer(async_mode='asgi')


@sio.event
def write(sid, line) -> None:
    serial.write(f'{augment(line)}\n'.encode())


async def receive() -> None:
    global buffer
    try:
        while not stop_requested:
            try:
                buffer += serial.read_all().decode()
                await sio.sleep(0)
            except UnicodeDecodeError as e:
                logging.exception('could not decode', exc_info=e)
            if '\n' in buffer:
                line, buffer = buffer.split('\n', 1)
                await sio.emit('read', check(line.rstrip('\r')))
    except Exception:
        logging.exception('could not read')


if __name__ == '__main__':
    serial = serial.Serial('/dev/ttyTHS1', 115200)
    buffer = ''
    app = socketio.ASGIApp(sio, on_startup=lambda: sio.start_background_task(receive))
    stop_requested = False
    try:
        uvicorn.run(app, host='0.0.0.0', port=8081)
    except KeyboardInterrupt:
        stop_requested = True



from pathlib import Path

from nicegui import app, ui
from starlette import responses

PATH = Path('~/.rosys/timelapse/videos').expanduser()


class VideosPage:

    def __init__(self) -> None:
        @ui.page('/videos', title='Videos')
        def page():
            def update_list() -> None:
                list.clear()
                with list:
                    for mp4 in sorted(PATH.glob('*.mp4'), reverse=True):
                        ui.link(mp4.stem.replace('_', ' ').replace('-', ':'), f'/timelapse/{mp4.name}')
            list = ui.column()
            ui.timer(5, update_list)

        @app.get('/timelapse/{filename}')
        def get_timelapse(filename: str) -> responses.FileResponse:
            return responses.FileResponse(PATH / filename)



from functools import wraps
from typing import Optional

from line_profiler import LineProfiler as PyUtilsLineProfiler


class LineProfiler:

    def __init__(self):
        self.functions: list[list] = []
        self.line_profiler: Optional[PyUtilsLineProfiler] = None

    def __call__(self, func):
        index = len(self.functions)

        @wraps(func)
        def wrap(*args, **kw):
            return self.functions[index][1](*args, **kw)

        self.functions.append([func, func])
        return wrap

    def start(self) -> None:
        self.line_profiler = PyUtilsLineProfiler()
        for f in self.functions:
            f[1] = self.line_profiler(f[0])

    def stop(self, *, print: bool = True) -> None:
        for f in self.functions:
            f[1] = f[0]
        if print:
            self.print()

    def print(self) -> None:
        if self.line_profiler:
            self.line_profiler.print_stats()

    def reset(self) -> None:
        self.stop(print=False)
        self.start()


profile = LineProfiler()



from nicegui import ui
from nicegui.elements.chart import Chart

from ...rosys import config


class LizardStats(Chart):

    def __init__(self) -> None:
        super().__init__({
            'title': {'text': 'Lizard', 'floating': True, 'y': 20},
            'chart': {'type': 'line', 'animation': False},
            'xAxis': {'labels': False},
            'yAxis': {'min': 0, 'title': {'text': 'ms'}},
            'series': [
                {'name': 'responsiveness', 'data': []},
                {'name': 'update', 'data': []},
                {'name': 'processing', 'data': []},
            ],
            'plotOptions': {'series': {'marker': False}},
            'navigation': {'buttonOptions': {'enabled': False}},
            'credits': False,
        })
        ui.timer(config.ui_update_interval, self.update)

    def update(self) -> None:
        self.options.series[0].data.append(max(self.lizard.responsiveness_stats or [0]) * 1000)
        self.options.series[1].data.append(max(self.lizard.update_stats or [0]) * 1000)
        self.options.series[2].data.append(max(self.lizard.processing_stats or [0]) * 1000)
        [s.data.pop(0) for s in self.options.series if len(s.data) > 20]



from .module import ModuleHardware
from .robot_brain import RobotBrain
from .serial import SerialHardware


class ExpanderHardware(ModuleHardware):
    """The expander module represents a second ESP microcontroller connected to the core ESP via serial."""

    def __init__(self, robot_brain: RobotBrain, *,
                 name: str = 'p0',
                 serial: SerialHardware,
                 boot: int = 25,
                 enable: int = 14) -> None:
        self.name = name
        lizard_code = f'{name} = Expander({serial.name}, {boot}, {enable})'
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code)



import abc
from typing import Callable, Optional

import numpy as np

from .. import helpers, rosys
from ..event import Event
from ..helpers import remove_indentation
from .bms_message import BmsMessage
from .bms_state import BmsState
from .expander import ExpanderHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Bms(Module, abc.ABC):
    """The BMS module communicates with a simple battery management system over a serial connection.

    The BMS module provides measured voltages as an event.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.state = BmsState()

        self.VOLTAGE_MEASURED = Event()
        """new voltage measurements are available for processing (argument: list of voltages)"""


class BmsHardware(Bms, ModuleHardware):
    """This module implements the hardware interface for the BMS module."""

    UPDATE_INTERVAL = 5.0

    def __init__(self, robot_brain: RobotBrain, *,
                 expander: Optional[ExpanderHardware] = None,
                 name: str = 'bms',
                 rx_pin: int = 26,
                 tx_pin: int = 27,
                 baud: int = 9600,
                 num: int = 1) -> None:
        self.name = name
        self.expander = expander
        lizard_code = remove_indentation(f'''
            {name} = {expander.name + "." if expander else ""}Serial({rx_pin}, {tx_pin}, {baud}, {num})
            {name}.unmute()
        ''')
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code)
        rosys.on_repeat(self._request, 1.0)
        self.message_hooks[name] = self._handle_bms

    async def _request(self) -> None:
        if rosys.time() > self.state.last_update + self.UPDATE_INTERVAL:
            await self.robot_brain.send(f'{self.name}.send(0xdd, 0xa5, 0x03, 0x00, 0xff, 0xfd, 0x77)')

    def _handle_bms(self, line: str) -> None:
        skip = 2 if self.expander else 1
        words = line.split()[skip:]
        msg = BmsMessage([int(w, 16) for w in words])
        msg.check()
        result = msg.interpret()
        self.state.percentage = result.get('capacity percent')
        self.state.voltage = result.get('total voltage')
        self.state.current = result.get('current')
        self.state.temperature = np.mean(result['temperatures']) if 'temperatures' in result else None
        self.state.is_charging = (self.state.current or 0) > -0.4
        self.state.last_update = rosys.time()


class BmsSimulation(Bms, ModuleSimulation):
    """This module simulates a BMS module."""

    AVERAGE_VOLTAGE = 25.0
    VOLTAGE_AMPLITUDE = 1.0
    VOLTAGE_FREQUENCY = 0.01
    MIN_VOLTAGE = 22.5
    MAX_VOLTAGE = 27.5
    CHARGING_CURRENT = 1.0
    DISCHARGING_CURRENT = -0.7
    AVERAGE_TEMPERATURE = 20.0
    TEMPERATURE_AMPLITUDE = 1.0
    TEMPERATURE_FREQUENCY = 0.01

    def __init__(self, is_charging: Optional[Callable[[], bool]] = None, fixed_voltage: Optional[float] = None) -> None:
        super().__init__()
        self.is_charging = is_charging
        self.fixed_voltage = fixed_voltage

    async def step(self, dt: float) -> None:
        self.state.is_charging = self.is_charging is not None and self.is_charging()
        self.state.voltage = \
            self.AVERAGE_VOLTAGE + self.VOLTAGE_AMPLITUDE * np.sin(self.VOLTAGE_FREQUENCY * rosys.time()) \
            if self.fixed_voltage is None else self.fixed_voltage
        self.state.percentage = helpers.ramp(self.state.voltage, self.MIN_VOLTAGE, self.MAX_VOLTAGE, 0.0, 100.0)
        self.state.current = self.CHARGING_CURRENT if self.state.is_charging else self.DISCHARGING_CURRENT
        self.state.temperature = self.AVERAGE_TEMPERATURE + \
            self.TEMPERATURE_AMPLITUDE * np.sin(self.TEMPERATURE_FREQUENCY * rosys.time())
        self.state.last_update = rosys.time()



from .asyncio_warnings import AsyncioWarnings
from .kpi_buckets import Day, Month, TimeBucket, Week
from .kpi_logger import KpiLogger, date_to_str, str_to_date
from .kpi_page_ import kpi_page
from .legacy.asyncio_monitor import AsyncioMonitor
from .legacy.network_monitor import NetworkMonitor, NetworkStats
from .legacy.objgraph_page import objgraph_page
from .memory import MemoryMiddleware
from .profile_button_ import ProfileButton as profile_button
from .timelapse_recorder import TimelapseRecorder
from .track import track
from .videos_page_ import VideosPage as videos_page



#!/usr/bin/env python3
import time

import numpy as np
import pylab as pl
from nicegui import ui
from rosys.geometry import Pose, Spline
from rosys.pathplanning import plot_tools as pt
from rosys.pathplanning.grid import Grid
from rosys.pathplanning.obstacle_map import ObstacleMap
from rosys.pathplanning.robot_renderer import RobotRenderer

grid = Grid((30, 40, 36), (0.45, -0.05, 4.0, 3.0))
obstacles = [
    [0.5, 1.5, 1.4, 0.1],
    [2.7, 1.5, 1.8, 0.1],
]
robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)

t = time.time()
obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)
ui.label('%.3f ms' % ((time.time() - t) * 1000))

start = [1.0, 0.5, 0]
end = [2.3, 0.9, np.deg2rad(90)]
spline = Spline.from_poses(
    Pose(x=start[0], y=start[1], yaw=start[2]),
    Pose(x=end[0], y=end[1], yaw=end[2]),
)

with ui.plot():
    pt.show_obstacle_map(obstacle_map)
    pl.autoscale(False)
    pt.plot_robot(robot_renderer, start, 'C3' if obstacle_map.test(*start) else 'C2')
    pt.plot_robot(robot_renderer, end, 'C3' if obstacle_map.test(*end) else 'C2')
    pt.plot_spline(spline, 'C3' if obstacle_map.test_spline(spline) else 'C2')

with ui.plot():
    pl.imshow(obstacle_map.dist_stack[:, :, 9], cmap=pl.cm.gray)

ui.run()
